package com.interactivemesh.j3d.community.gui;

/*
 * LICENSE
 * 
 * FXCanvas3DImage API 2.0
 * 
 * This API is a derived work of the Java 3D class "com.sun.j3d.exp.swing.JCanvas3D.java"
 * 
 * and includes the files
 * 
 * - "com.interactivemesh.j3d.community.gui.FXCanvas3DImage.java"
 * - "com.interactivemesh.j3d.community.gui.FXCanvas3DImageRepainter.java" 
 * 
 * Redistribution and use are permitted according to the following license notice.
 * 
 * Version: 2.0
 * Date: 2010/05/14
 * 
 * Author:
 * August Lammersdorf, InteractiveMesh e.K.
 * Kolomanstrasse 2a, 85737 Ismaning 
 * Germany / Munich Area
 * www.InteractiveMesh.com/org
 * 
*/
/*
 * $RCSfile: com.sun.j3d.exp.swing.JCanvas3D.java,v $
 *
 * Copyright (c) 2007 Sun Microsystems, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistribution of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistribution in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * Neither the name of Sun Microsystems, Inc. or the names of
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * This software is provided "AS IS," without a warranty of any
 * kind. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY
 * EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL
 * NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
 * USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES. IN NO EVENT WILL SUN OR ITS LICENSORS BE LIABLE FOR
 * ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
 * CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
 * REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
 * INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES.
 *
 * You acknowledge that this software is not designed, licensed or
 * intended for use in the design, construction, operation or
 * maintenance of any nuclear facility.
 *
 * $Revision: 1.10 $
 * $Date: 2007/04/11 02:08:56 $
 * $State: Exp $
 * 
 *  @author: Frederic 'pepe' Barachant
 *
 */

import java.awt.AWTEvent;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.GraphicsConfigTemplate;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Rectangle;

import java.awt.event.ComponentEvent;
import java.awt.event.FocusEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;

import java.awt.image.BufferedImage;
import java.awt.image.DataBufferInt;

import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

import javax.media.j3d.Canvas3D;
import javax.media.j3d.GraphicsConfigTemplate3D;
import javax.media.j3d.ImageComponent2D;
import javax.media.j3d.RestrictedAccessException;
import javax.media.j3d.Screen3D;

import javax.swing.JPanel;
import javax.swing.SwingUtilities;

import javax.swing.event.AncestorEvent;
import javax.swing.event.AncestorListener;

import com.sun.j3d.exp.swing.impl.AutoOffScreenCanvas3D;

/**
 * <p>
 * This class provides Java 3D rendering into a BufferedImage object 
 * which can be drawn within a JavaFX Scene. 
 * </p> 
 * <p>The orientation of the image data is "y-up" !
 * </p> 
 *
 * @since Java 3D 1.5
 * 
 * @see FXCanvas3DImageRepainter
 */
public class FXCanvas3DImage extends JPanel implements AncestorListener {  

    /** size of a pixel */
    private static double METERS_PER_PIXEL = 0.0254 / 90.0;
    
    /** the graphics configuration used for this canvas */
    private GraphicsConfiguration   graphicsConfig  =   null;  

    /** The canvas that is linked to the component. */
    private OffscreenCanvas3D       canvas          =   null;
        
    private long                    wakeupEventMasks =  0;

    //
    // Double Offscreen Buffer
    //   
    private BufferedImage       paintImage          =   null;
    //
    // Lock - Conditions
    //
    private ReentrantLock       imageAccessLock     =   null;
    private Condition           imageDrawnCondition =   null;
    private Condition           imageResizeCondition =  null;
    private volatile boolean    isImageDrawn       	=   false;
    private volatile boolean    isRendererLocked    =   false;
    private boolean             isSwapStarted       =   false;  
    //
    // FXCanvas3DImage Repainter
    //
    private FXCanvas3DImageRepainter repainter      =   null;

    /**
     * Constructs and initializes a new FXCanvas3DImage object. 
     * The screen device is obtained from
     * <code>GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice()</code>,
     * which might not be the one you should use if you are in a multiscreen environment.
     * The FXCanvas3DImage is constructed using the following default parameters:<br><br>
     * double buffer enable : false<br>
     * stereo enable : false<br>
     */
    public FXCanvas3DImage() {
        this(null, null);
    }

    /**
     * Constructs and initializes a new FXCanvas3DImage object, using the specified graphics device.
     *
     * @param device the screen graphics device that will be used to construct
     *        a GraphicsConfiguration.
     */
    public FXCanvas3DImage(GraphicsDevice device) {
        this(null, device);
    }

    /**
     * Constructs and initializes a new FXCanvas3DImage object, using the specified template.
     * The screen device is obtained from
     * <code>GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice()</code>,
     * which might not be the one you should use if you are
     * in a multiscreen environment.
     *
     * @param template The template that will be used to construct a
     *        GraphicsConfiguration. The stereo and doublebuffer properties
     *        are forced to UNNECESSARY.
     */
    public FXCanvas3DImage(GraphicsConfigTemplate3D template) {
        this(template, null);
    }

    /**
     * Constructs and initializes a new FXCanvas3DImage object, 
     * using the specified template and graphics device.
     *
     * @param template The template that will be used to construct a
     *        GraphicsConfiguration. The stereo and doublebuffer properties
     *        are forced to UNNECESSARY.
     * @param device the screen graphics device that will be used to construct
     *        a GraphicsConfiguration in conjunction with the template.
     */
    public FXCanvas3DImage(GraphicsConfigTemplate3D template, GraphicsDevice device) {
        
        if (template == null) {
            template = new GraphicsConfigTemplate3D();
        }
        // Force double-buffer and stereo to UNNECESSARY
        template.setStereo(GraphicsConfigTemplate.UNNECESSARY);
        template.setDoubleBuffer(GraphicsConfigTemplate.UNNECESSARY);

        if (device == null) {
            graphicsConfig = GraphicsEnvironment.getLocalGraphicsEnvironment().
                                getDefaultScreenDevice().getBestConfiguration(template);
        }
        else {       
            graphicsConfig = device.getBestConfiguration(template);
        }

        this.addAncestorListener(this);
        this.setDoubleBuffered(false);

        // so that key events can be received.
        this.setFocusable(true);
        // A ComponentListener is added to the offscreen Canvas3D by default !
        this.enableEvents(AWTEvent.COMPONENT_EVENT_MASK);

        // Sync image access
        imageAccessLock      = new ReentrantLock();
        imageDrawnCondition  = imageAccessLock.newCondition();
        imageResizeCondition = imageAccessLock.newCondition();
    }
    
    //
    // JavaFX - Java 3D bridge
    //
    
    /** 
     * Sets the FXCanvas3DImageRepainter object for this FXCanvas3DImage.
     * 
     * @param painter JavaFX class instance
     */
    public void setRepainter(FXCanvas3DImageRepainter painter) {
        repainter = painter;
    }
       
    /**
     * Returns the BufferedImage object that contains the 3D rendering.
     * The image data will be updated per rendering loop. 
     * 
     * @return the image of the 3D rendering 
     * @see #copyOffScreenBuffer()
     */
    public BufferedImage getCanvas3DImage() {       
        if (this.canvas == null) {
            createCanvas(getWidth(), getHeight());
        }
        return paintImage;
    }
    
    /** 
     * Notifies the waiting J3D-Renderer thread.
     * To be called from the JavaFX painting loop when the repainting of the image is done.
     * 
     * @since 2.0
    */
	public void imageDrawn() {
	    try {      
	        imageAccessLock.lock();
	        // Notify the waiting J3D-Renderer thread.
	        isImageDrawn = true;
	        imageDrawnCondition.signal();       
	    }
	    finally {
	        imageAccessLock.unlock();
	    }
	}
    
    //
    // Interface AncestorListener
    //
    
    /**
     * {@inheritDoc}
     *
     * @param event {@inheritDoc}
     */
     public void ancestorAdded(AncestorEvent event) {
        Dimension sz = getSize();

        if (sz.width < 1) {
            sz.width = 100;
        }

        if (sz.height < 1) {
            sz.height = 100;
        }

        createCanvas(sz.width, sz.height);

        canvas.addNotifyFlag = true; // make it so that i can call addNotify() without being rejected.
        canvas.addNotify();
    }

    /** 
     * {@inheritDoc}
     *
     * @param event {@inheritDoc}
     */
    public void ancestorMoved(AncestorEvent event) {
    }

    /**
     * {@inheritDoc}
     *
     * @param event {@inheritDoc}
     */
    public void ancestorRemoved(AncestorEvent event) {
        canvas.removeNotify();
    }

    //
    // Java 3D specific methods
    //
    
    /**
     * Returns the off-screen heavyweight canvas of this lightweight component.
     *
     * @return the heavyweight canvas that lies in the deepness of this
     *         Component.
     */
    public Canvas3D getOffscreenCanvas3D() {
        if (this.canvas == null) {
            createCanvas(getWidth(), getHeight());
        }

        return this.canvas;
    }
    
    /**
     * {@inheritDoc}
     *
     * @param x {@inheritDoc}
     * @param y {@inheritDoc}
     * @param width {@inheritDoc}
     * @param height {@inheritDoc}
     */
    @Override
    public void setBounds(int x, int y, int width, int height) {

        super.setBounds(x, y, width, height);

        createCanvas(width, height);                   
    }

    /**
     * Creates a heavyweight canvas and initializes it, or changes the
     * size of the current one if present. Current heavyweight canvas is
     * changed only if size is different from the actual one. 
     * 
     * No canvas is created if this component has no parent, that is, 
     * was not added to a container.
     *
     * @param width the width of the canvas to create.
     * @param height the height of the canvas to create.
     */
    private void createCanvas(int width, int height) {
       
        if (getParent() == null) {
            return;
        }
        
        final int w = Math.max(1, width);
        final int h = Math.max(1, height);
        
        if (canvas == null) {
            canvas = new OffscreenCanvas3D(this.graphicsConfig);
        } 
        else {
            if (canvas.getWidth() == w && canvas.getHeight() == h) 
                return;
        }

        // Create 'OffScreen Buffer'

        if (SwingUtilities.isEventDispatchThread()) {
            canvas.createOffScreenBuffer(graphicsConfig.getBounds(), w, h);
        }
        else {
            SwingUtilities.invokeLater(new Runnable() { 
                public void run() {
                    canvas.createOffScreenBuffer(graphicsConfig.getBounds(), w, h);
                }
            });
        }       
    }
    
    /* 
    @Override
    public void paintComponent(Graphics g) {        
        super.paintComponent(g); // paint background        
    }*/

    
    // WakeupOnAWTEvent: COMPONENT_EVENT_MASK, FOCUS_EVENT_MASK, 
    //                   KEY_EVENT_MASK
    //                   MOUSE_EVENT_MASK, MOUSE_MOTION_EVENT_MASK, MOUSE_WHEEL_EVENT_MASK
    /**
     * Sets the AWT event masks for which this FXCanvas3DImage object 
     * will call the corresponding process-methods on the underlying heavyweight canvas.
     * <p>
     * This is required only when Java 3D WakeupOnAWTEvents are specified in subclasses of Behavior. 
     * <p>
     * No event mask is set per default.
     * 
     * @param eventMasks ored AWT event masks: <code>AWTEvent.COMPONENT_EVENT_MASK,
     * AWTEvent.FOCUS_EVENT_MASK, AWTEvent.KEY_EVENT_MASK, AWTEvent.MOUSE_EVENT_MASK,
     * AWTEvent.MOUSE_MOTION_EVENT_MASK, AWTEvent.MOUSE_WHEEL_EVENT_MASK</code>.
     * 
     * @since 1.1
     */
    public void enableWakeupOnAWTEvents(long eventMasks) {
        
        if ((eventMasks & AWTEvent.COMPONENT_EVENT_MASK) != 0)
            wakeupEventMasks |= AWTEvent.COMPONENT_EVENT_MASK;
        if ((eventMasks & AWTEvent.FOCUS_EVENT_MASK) != 0)
            wakeupEventMasks |= AWTEvent.FOCUS_EVENT_MASK;
        if ((eventMasks & AWTEvent.KEY_EVENT_MASK) != 0)
            wakeupEventMasks |= AWTEvent.KEY_EVENT_MASK;
        if ((eventMasks & AWTEvent.MOUSE_EVENT_MASK) != 0)
            wakeupEventMasks |= AWTEvent.MOUSE_EVENT_MASK;
        if ((eventMasks & AWTEvent.MOUSE_MOTION_EVENT_MASK) != 0)
            wakeupEventMasks |= AWTEvent.MOUSE_MOTION_EVENT_MASK;
        if ((eventMasks & AWTEvent.MOUSE_WHEEL_EVENT_MASK) != 0)
            wakeupEventMasks |= AWTEvent.MOUSE_WHEEL_EVENT_MASK;        
        /*
        Enables the events defined by the specified event mask parameter 
        to be delivered to this component.

        Event types are automatically enabled when a listener for that event type 
        is added to the component.

        This method only needs to be invoked by subclasses of Component 
        which desire to have the specified event types delivered 
        to processEvent regardless of whether or not a listener is registered. 
        */
        this.enableEvents(wakeupEventMasks);
    }
    /**
     * Disables calling event process-methods on the underlying heavyweight canvas
     * for the specified AWT event masks.
     * @param eventMasks ored AWT event masks: <code>AWTEvent.COMPONENT_EVENT_MASK,
     * AWTEvent.FOCUS_EVENT_MASK, AWTEvent.KEY_EVENT_MASK, AWTEvent.MOUSE_EVENT_MASK,
     * AWTEvent.MOUSE_MOTION_EVENT_MASK, AWTEvent.MOUSE_WHEEL_EVENT_MASK</code>
     * 
     * @since 1.1
     */
    public void disableWakeupOnAWTEvents(long eventMasks) {
        wakeupEventMasks &= ~eventMasks;
        this.disableEvents(eventMasks);
    }
    
    //
    // Override Component's processXXX
    //
    /**
     * Redirects event to canvas and to superclass.
     *
     * @param e {@inheritDoc}
     */
    @Override
    protected void processComponentEvent(ComponentEvent e) {
        super.processComponentEvent(e);
      
//        if ((wakeupEventMasks & AWTEvent.COMPONENT_EVENT_MASK) != 0) {
            Object src = e.getSource();
            e.setSource(canvas);
            canvas.processComponentEvent(e);
            e.setSource(src);
//        }
    }

    /**
     * Redirects event to canvas and to superclass.
     *
     * @param e {@inheritDoc}
     */
    @Override
    protected void processFocusEvent(FocusEvent e) {
        super.processFocusEvent(e);

        if ((wakeupEventMasks & AWTEvent.FOCUS_EVENT_MASK) != 0) {
            Object src = e.getSource();
            e.setSource(canvas);
            canvas.processFocusEvent(e);
            e.setSource(src);
        }
    }

    /**
     * Redirects event to canvas and to superclass.
     *
     * @param e {@inheritDoc}
     */
    @Override
    protected void processKeyEvent(KeyEvent e) {
        super.processKeyEvent(e);

        if ((wakeupEventMasks & AWTEvent.KEY_EVENT_MASK) != 0) {
            Object src = e.getSource();
            e.setSource(canvas);
            canvas.processKeyEvent(e);
            e.setSource(src);
        }
    }

    /**
     * Redirects event to canvas and to superclass.
     *
     * @param e {@inheritDoc}
     */
    @Override
    protected void processMouseEvent(MouseEvent e) {
        super.processMouseEvent(e);

        if ((wakeupEventMasks & AWTEvent.MOUSE_EVENT_MASK) != 0) {
            Object src = e.getSource();
            e.setSource(canvas);
            canvas.processMouseEvent(e);
            e.setSource(src);
        }
    }

    /**
     * Redirects event to canvas and to superclass.
     *
     * @param e {@inheritDoc}
     */
    @Override
    protected void processMouseMotionEvent(MouseEvent e) {
        super.processMouseMotionEvent(e);

        if ((wakeupEventMasks & AWTEvent.MOUSE_MOTION_EVENT_MASK) != 0) {
            Object src = e.getSource();
            e.setSource(canvas);
            canvas.processMouseMotionEvent(e);
            e.setSource(src);
        }
    }

    /**
     * Redirects event to canvas and to superclass.
     *
     * @param e {@inheritDoc}
     */
    @Override
    protected void processMouseWheelEvent(MouseWheelEvent e) {
        super.processMouseWheelEvent(e);

        if ((wakeupEventMasks & AWTEvent.MOUSE_WHEEL_EVENT_MASK) != 0) {
            Object src = e.getSource();
            e.setSource(canvas);
            canvas.processMouseWheelEvent(e);
            e.setSource(src);
        }
    }

    /**
     * This class is the internal Canvas3D that is used and sent to
     * Java 3D. It is remote controlled through FXCanvas3DImage and is modified to be
     * able to tell the lightweight component when refreshes are needed.
     */
    private final class OffscreenCanvas3D extends    Canvas3D 
                                          implements AutoOffScreenCanvas3D {

        // Render target
        private volatile boolean isWatingForResizing = false;  

        /**
         * Flag used to sort a call to addnotify() from user and
         * from the lightweight component. Lightweight component calls
         * addNotify() so that the rendering begins and uses normal routines,
         * but this is a method that user must not call.
         */
        private boolean addNotifyFlag = false;
        
        // Call repaint on EDT
        private final Runnable repaintCall = new Runnable() {
            public void run() {
                repainter.repaintFXCanvas3DImage();
            }
        };
        
        // Don't wait longer than 200 ms for repainting
        private static final int LIMIT = 200000000;
        
        /**
         * Creates a new instance of InternalCanvas3D.
         * 
         * @param graphicsConfiguration The graphics configuration to be used.
         * @param lwCanvas the lightweight canvas that is linked to that
         *        heavyweight canvas.
         */
        private OffscreenCanvas3D(GraphicsConfiguration graphicsConfiguration) {
            super(graphicsConfiguration, true);
        }
        
        // To be called on the EDT only !!!!!!!!!!!!!!
        /**
         * Creates an offscreen buffer to be attached to the heavyweight
         * buffer. Buffer is created 'byreference'
         *
         * @param width the width of the buffer.
         * @param height the height of the buffer.
         */
        private void createOffScreenBuffer(Rectangle screenRect, int width, int height) {

            try {
                imageAccessLock.lock();
                
	            // Now the J3D-Renderer thread can't enter the repaint code block in postSwap
	            this.stopRenderer(); 
	
	            // Compute physical dimensions of the screen
	            int screenWidth = (int)screenRect.getWidth();
	            int screenHeight = (int)screenRect.getHeight();
	            Screen3D screen3D = this.getScreen3D();
	            screen3D.setSize(screenWidth, screenHeight);
	            screen3D.setPhysicalScreenWidth(((double) screenWidth) * METERS_PER_PIXEL);
	            screen3D.setPhysicalScreenHeight(((double) screenHeight) * METERS_PER_PIXEL);
	            
	            if (paintImage != null)
	                paintImage.flush();
	            
	            // OffScreenBuffer: byReference & yUp
	            paintImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
	            ImageComponent2D image = new ImageComponent2D(ImageComponent2D.FORMAT_RGBA, paintImage, true, true);
	            image.setCapability(ImageComponent2D.ALLOW_IMAGE_READ);
	            image.setCapability(ImageComponent2D.ALLOW_IMAGE_WRITE);
	            
            	isWatingForResizing = true;
            	
                // Setting offscreen buffer requires that the J3D-Renderer thread isn't blocked
                // As we are on the EDT, FXCanvas3DImageRepainter can't release the J3D-Renderer
                // So, it's done here. We are waiting until this has happened.
                while (isRendererLocked) {
                    
                	isImageDrawn = true;
                	imageDrawnCondition.signal();
                     
                     try {
                        imageResizeCondition.await();
                     }
                     catch (InterruptedException e) {
                     }
                 }
                
                isWatingForResizing = false;
                
                // Offscreen rendering might occur even if the renderer is stopped. 
                // For that reason, we are waiting for a hypothetical offscreen render 
                // to finish before setting offscreen rendering.
                // Otherwise, rendering will stop with an exception.
                this.waitForOffScreenRendering();               

                try {
                    // EDT will wait (sleep) until exchange of offscreen buffer is done 
                    this.setOffScreenBuffer(image);                
                }
                catch (RestrictedAccessException e) {
                    // Try again! Risk of endless loop ?  
                    createOffScreenBuffer(screenRect, width, height); 
                    System.out.println("Repeat : createOffScreenBuffer(screenRect, width, height) !!");
                }
            }
            finally  {
                imageAccessLock.unlock();
            }
                       
            // Callback here ok ??
            if (repainter != null) {
                repainter.onFXCanvas3DImageResized();
            }

            this.startRenderer();
        }
        
        /**
         * {@inheritDoc}
         */
        @Override
        public void addNotify() {
            if (addNotifyFlag) {
                addNotifyFlag = false;
                super.addNotify();                
            } 
            else {
                throw new UnsupportedOperationException("User must not call this method !!");
            }
        }

        /**
         * Normally, returns the parent of that component. As the
         * canvas ought to never be added to any component, it has no parent.
         * Java 3D expects it to have a parent for some operations, so we in
         * fact cheat it by returning the parent of the lightweight component.
         *
         * @return the parent of the lightweight component, if any. Returns
         *         null if it has no parent.
         */
        @Override 
        public Container getParent() {
            return FXCanvas3DImage.this.getParent();
        }

        @Override
        public void preRender() {
            if (isSwapStarted) {
                isSwapStarted = false;
                try {
                    imageAccessLock.unlock();                    
                }
                catch (IllegalMonitorStateException e) {                    
                }
            }
        }
        @Override
        public void postRender() {
            if (this.isRendererRunning()) {             
                imageAccessLock.lock();              
                isSwapStarted = true;          
            }
        }
        @Override
        public void postSwap() {
                   
            if (isSwapStarted) { 
            		            	
            	isImageDrawn = false;
                
                // Repainting in JavaFX painting loop on EDT                    
                SwingUtilities.invokeLater(repaintCall);
                
                // Wait until image is drawn
                while (!isImageDrawn) { 
                    
                    isRendererLocked = true;
                    
                    try {                           
                    	imageDrawnCondition.awaitNanos(LIMIT);   // Don't wait for ever
                    	isImageDrawn = true;                     // and release yourself
                    }
                    catch (InterruptedException e) {
                    }
                    
                    isRendererLocked = false;
                    if (isWatingForResizing)
                    	imageResizeCondition.signal();
                }
                
                isSwapStarted = false;
                
                try  {                      
                   imageAccessLock.unlock();
                }               
                catch (IllegalMonitorStateException e) {                    
                }            
            }
        }
        
        // ----------------- Java 3D engine ------------------------------------
        
        // WakeupOnAWTEvent: COMPONENT_EVENT_MASK, FOCUS_EVENT_MASK, 
        //                   KEY_EVENT_MASK
        //                   MOUSE_EVENT_MASK, MOUSE_MOTION_EVENT_MASK, MOUSE_WHEEL_EVENT_MASK

        // Canvas3D / WakeupOnAWTEvent
        // EventCatcher: FocusListener, KeyListener, 
        //               MouseListener, MouseMotionListener, MouseWheelListener
        
        // Canvas3D, all ancestors, Window
        // EventCatcher: ComponentListener
        
        // Canvas3D, all ancestors, Window
        // CanvasViewEventCatcher: ComponentAdapter (componentResized/componentMoved) 

        // Window
        // EventCatcher: WindowListener
        
        // Hint: Offscreen-Canvas3D: componentMoved, componentHidden, componentShown
        //                           are not relevant
        //                           componentResized is called due to 
        //                                            setSize-call in setOffScreenBuffer
        
        // ---------------------------------------------------------------------

        //
        // Override Component's processXXX
        //
        /**
         * Overriden so that the JComponent can access it.
         *
         * @param e {@inheritDoc}
         */
        @Override
        protected void processComponentEvent(ComponentEvent e) {
            super.processComponentEvent(e);
        }

        /**
         * Overriden so that the JComponent can access it.
         *
         * @param e {@inheritDoc}
         */
        @Override
        protected void processFocusEvent(FocusEvent e) {
            super.processFocusEvent(e);
        }

        /**
         * Overriden so that the JComponent can access it.
         *
         * @param e {@inheritDoc}
         */
        @Override
        protected void processKeyEvent(KeyEvent e) {
            super.processKeyEvent(e);
        }

        /**
         * Overriden so that the JComponent can access it.
         *
         * @param e {@inheritDoc}
         */
        @Override
        protected void processMouseEvent(MouseEvent e) {
            super.processMouseEvent(e);
        }

        /**
         * Overriden so that the JComponent can access it.
         *
         * @param e {@inheritDoc}
         */
        @Override
        protected void processMouseMotionEvent(MouseEvent e) {
            super.processMouseMotionEvent(e);
        }

        /**
         * Overriden so that the JComponent can access it.
         *
         * @param e {@inheritDoc}
         */
        @Override
        protected void processMouseWheelEvent(MouseWheelEvent e) {
            super.processMouseWheelEvent(e);
        }
    }
}
